import random
import time
from .mouse import Mouse
from .connection import SerialTransport
from .errors import MakcuConnectionError
from .enums import MouseButton

class MakcuController:
    def __init__(self, fallback_com_port, debug=False, send_init=True):
        self.transport = SerialTransport(fallback_com_port, debug=debug, send_init=send_init)
        self.mouse = Mouse(self.transport)

    def connect(self):
        self.transport.connect()

    def disconnect(self):
        self.transport.disconnect()

    def is_connected(self):
        return self.transport.is_connected()

    def _check_connection(self):
        if not self.transport.serial or not self.transport.serial.is_open:
            raise MakcuConnectionError("Not connected")

    def click(self, button: MouseButton):
        self._check_connection()
        self.mouse.press(button)
        self.mouse.release(button)

    def move(self, dx: int, dy: int):
        self._check_connection()
        self.mouse.move(dx, dy)

    def scroll(self, delta: int):
        self._check_connection()
        self.mouse.scroll(delta)

    def move_smooth(self, dx: int, dy: int, segments: int):
        self._check_connection()
        self.mouse.move_smooth(dx, dy, segments)

    def move_bezier(self, dx: int, dy: int, segments: int, ctrl_x: int, ctrl_y: int):
        self._check_connection()
        self.mouse.move_bezier(dx, dy, segments, ctrl_x, ctrl_y)

    def lock_mouse_x(self, lock: bool):
        self._check_connection()
        self.mouse.lock_x(lock)

    def lock_mouse_y(self, lock: bool):
        self._check_connection()
        self.mouse.lock_y(lock)

    def lock_left(self, lock: bool):
        self._check_connection()
        self.mouse.lock_left(lock)

    def lock_middle(self, lock: bool):
        self._check_connection()
        self.mouse.lock_middle(lock)

    def lock_right(self, lock: bool):
        self._check_connection()
        self.mouse.lock_right(lock)

    def lock_side1(self, lock: bool):
        self._check_connection()
        self.mouse.lock_side1(lock)

    def lock_side2(self, lock: bool):
        self._check_connection()
        self.mouse.lock_side2(lock)

    def spoof_serial(self, serial: str):
        self._check_connection()
        self.mouse.spoof_serial(serial)

    def reset_serial(self):
        self._check_connection()
        self.mouse.reset_serial()

    def get_device_info(self):
        self._check_connection()
        return self.mouse.get_device_info()

    def get_firmware_version(self):
        self._check_connection()
        return self.mouse.get_firmware_version()

    def get_button_mask(self) -> int:
        self._check_connection()
        return self.transport.get_button_mask()

    def is_button_locked(self, button: MouseButton) -> bool:
        self._check_connection()
        return self.mouse.is_button_locked(button)

    def capture(self, button: MouseButton):
        self._check_connection()
        self.mouse.begin_capture(button.name)

    def get_captured_clicks(self, button: MouseButton) -> int:
        self._check_connection()
        return self.mouse.stop_capturing_clicks(button.name)


    def click_human_like(self, button: MouseButton, count: int = 1,
        profile: str = "normal", jitter: int = 0):
        self._check_connection()

        timing_profiles = {
            "normal": {"min_down": 60,  "max_down": 120, "min_wait": 100, "max_wait": 180},
            "fast":   {"min_down": 30,  "max_down": 60,  "min_wait": 50,  "max_wait": 100},
            "slow":   {"min_down": 100, "max_down": 180, "min_wait": 150, "max_wait": 300},
        }

        if profile not in timing_profiles:
            raise ValueError(f"Invalid profile: {profile}. Choose from {list(timing_profiles.keys())}")

        t = timing_profiles[profile]

        for _ in range(count):
            if jitter > 0:
                dx = random.randint(-jitter, jitter)
                dy = random.randint(-jitter, jitter)
                self.mouse.move(dx, dy)

            self.mouse.press(button)
            time.sleep(random.uniform(t["min_down"], t["max_down"]) / 1000.0)
            self.mouse.release(button)
            time.sleep(random.uniform(t["min_wait"], t["max_wait"]) / 1000.0)
    
    def enable_button_monitoring(self, enable: bool = True):
        self._check_connection()
        self.transport.enable_button_monitoring(enable)

    def set_button_callback(self, callback):
        self._check_connection()
        self.transport.set_button_callback(callback)

    def get_all_lock_states(self) -> dict:
        self._check_connection()
        return self.mouse.get_all_lock_states()

    def stop_capturing_clicks(self, button: str) -> int:
        self._check_connection()
        return self.mouse.stop_capturing_clicks(button)

    def press(self, button: MouseButton):
        self._check_connection()
        self.mouse.press(button)

    def release(self, button: MouseButton):
        self._check_connection()
        self.mouse.release(button)
        
    def get_button_states(self) -> dict:
        self._check_connection()
        return self.transport.get_button_states()

    def is_button_pressed(self, button: MouseButton) -> bool:
        self._check_connection()
        return self.transport.get_button_states().get(button.name.lower(), False)


from .enums import MouseButton
from .errors import MakcuCommandError
from serial.tools import list_ports

class Mouse:
    def __init__(self, transport):
        self.transport = transport

    def press(self, button: MouseButton):
        cmd = {
            MouseButton.LEFT: "km.left(1)",
            MouseButton.RIGHT: "km.right(1)",
            MouseButton.MIDDLE: "km.middle(1)"
        }.get(button)
        if cmd:
            self.transport.send_command(cmd)
        else:
            raise MakcuCommandError(f"Unsupported button for press(): {button}")

    def release(self, button: MouseButton):
        cmd = {
            MouseButton.LEFT: "km.left(0)",
            MouseButton.RIGHT: "km.right(0)",
            MouseButton.MIDDLE: "km.middle(0)"
        }.get(button)
        if cmd:
            self.transport.send_command(cmd)
        else:
            raise MakcuCommandError(f"Unsupported button for release(): {button}")

    def move(self, x: int, y: int):
        self.transport.send_command(f"km.move({x},{y})")

    def move_smooth(self, x: int, y: int, segments: int):
        self.transport.send_command(f"km.move({x},{y},{segments})")

    def move_bezier(self, x: int, y: int, segments: int, ctrl_x: int, ctrl_y: int):
        self.transport.send_command(f"km.move({x},{y},{segments},{ctrl_x},{ctrl_y})")

    def scroll(self, delta: int):
        self.transport.send_command(f"km.wheel({delta})")

    def lock_left(self, lock: bool): self.transport.send_command(f"km.lock_ml({int(lock)})")
    def lock_middle(self, lock: bool): self.transport.send_command(f"km.lock_mm({int(lock)})")
    def lock_right(self, lock: bool): self.transport.send_command(f"km.lock_mr({int(lock)})")
    def lock_side1(self, lock: bool): self.transport.send_command(f"km.lock_ms1({int(lock)})")
    def lock_side2(self, lock: bool): self.transport.send_command(f"km.lock_ms2({int(lock)})")
    def lock_x(self, lock: bool): self.transport.send_command(f"km.lock_mx({int(lock)})")
    def lock_y(self, lock: bool): self.transport.send_command(f"km.lock_my({int(lock)})")

    def spoof_serial(self, serial: str): self.transport.send_command(f"km.serial('{serial}')")
    def reset_serial(self): self.transport.send_command("km.serial(0)")

    def get_device_info(self) -> dict:
        port_name = self.transport.port
        is_connected = self.transport.is_connected()
        info = {
            "port": port_name,
            "description": "Unknown",
            "vid": "Unknown",
            "pid": "Unknown",
            "isConnected": is_connected
        }
        for port in list_ports.comports():
            if port.device == port_name:
                info.update({
                    "description": port.description,
                    "vid": hex(port.vid) if port.vid is not None else "Unknown",
                    "pid": hex(port.pid) if port.pid is not None else "Unknown"
                })
                break
        return info

    def get_firmware_version(self) -> str:
        return self.transport.send_command("km.version()", expect_response=True)

    def is_locked(self, target: str) -> bool:
        target = target.upper()
        commands = {
            "X": "km.lock_mx()",
            "Y": "km.lock_my()",
            "LEFT": "km.lock_ml()",
            "MIDDLE": "km.lock_mm()",
            "RIGHT": "km.lock_mr()",
            "MOUSE4": "km.lock_ms1()",
            "MOUSE5": "km.lock_ms2()",
        }
        if target not in commands:
            raise ValueError(f"Unsupported lock target: {target}")
        try:
            result = self.transport.send_command(commands[target], expect_response=True)
            return result.strip() == "1"
        except Exception:
            return False

    def is_button_locked(self, button: MouseButton) -> bool:
        name_map = {
            MouseButton.LEFT: "LEFT",
            MouseButton.RIGHT: "RIGHT",
            MouseButton.MIDDLE: "MIDDLE",
            MouseButton.MOUSE4: "MOUSE4",
            MouseButton.MOUSE5: "MOUSE5"
        }
        return self.is_locked(name_map[button])

    def begin_capture(self, button: str):
        """
        Assumes lock_<button>(1) has already been called.
        Sends catch_<button>(0) to begin capturing click cycles.
        """
        self.transport.catch_button(button)

    def stop_capturing_clicks(self, button: str) -> int:
        """
        Assumes lock_<button>(0) has already been called.
        Returns the total number of clicks since begin_capture.
        """
        return self.transport.read_captured_clicks(button)

    def get_all_lock_states(self) -> dict:
        return {
            target: self.is_locked(target)
            for target in ["X", "Y", "LEFT", "RIGHT", "MIDDLE", "MOUSE4", "MOUSE5"]
        }


import pytest
from makcu import create_controller
import time

@pytest.fixture(scope="session")
def makcu():
    ctrl = create_controller()
    yield ctrl
    ctrl.disconnect()
    time.sleep(0.2)

@pytest.fixture(autouse=True)
def ensure_clean_exit(makcu):
    yield
    makcu.mouse.lock_left(False)
    makcu.mouse.lock_right(False)
    makcu.mouse.lock_middle(False)
    makcu.mouse.lock_side1(False)
    makcu.mouse.lock_side2(False)
    makcu.mouse.lock_x(False)
    makcu.mouse.lock_y(False)
    makcu.enable_button_monitoring(False)


import pytest
from makcu import MouseButton

def test_is_button_pressed(makcu):
    assert makcu.is_button_pressed(MouseButton.LEFT) in [True, False]

def test_press_and_release(makcu):
    makcu.press(MouseButton.LEFT)
    makcu.release(MouseButton.LEFT)

def test_firmware_version(makcu):
    print("Getting firmware version...")
    version = makcu.mouse.get_firmware_version()
    print(f"Firmware version: {version}")
    assert version and len(version.strip()) > 0

def test_middle_click(makcu):
    makcu.mouse.press(MouseButton.MIDDLE)
    makcu.mouse.release(MouseButton.MIDDLE)

def test_device_info(makcu):
    print("Fetching device info...")
    info = makcu.mouse.get_device_info()
    print(f"Device Info: {info}")
    assert info.get("port")
    assert info.get("isConnected") is True

def test_port_connection(makcu):
    assert makcu.is_connected()

@pytest.mark.skip(reason="Capture test disabled until firmware supports tracking clicks from software input")
def test_capture_right_clicks(makcu):
    makcu.mouse.lock_right(True)
    assert makcu.mouse.is_button_locked(MouseButton.RIGHT)

    makcu.mouse.begin_capture("RIGHT")
    makcu.mouse.press(MouseButton.RIGHT)
    makcu.mouse.release(MouseButton.RIGHT)
    makcu.mouse.press(MouseButton.RIGHT)
    makcu.mouse.release(MouseButton.RIGHT)

    makcu.mouse.lock_right(False)
    assert not makcu.mouse.is_button_locked(MouseButton.RIGHT)

    count = makcu.mouse.stop_capturing_clicks("RIGHT")
    assert count >= 2, f"Expected >=2 captured clicks, got {count}"

def test_button_mask(makcu):
    print("Getting button mask...")
    mask = makcu.get_button_mask()
    print(f"Mask value: {mask}")
    assert isinstance(mask, int)

def test_get_button_states(makcu):
    states = makcu.get_button_states()
    assert isinstance(states, dict)
    for key in ['left', 'right', 'middle', 'mouse4', 'mouse5']:
        assert key in states

def test_lock_state(makcu):
    print("Locking LEFT button...")
    makcu.lock_left(True)

    print("Querying lock state while LEFT is locked...")
    assert makcu.is_button_locked(MouseButton.LEFT)

    print("Querying all lock states...")
    all_states = makcu.get_all_lock_states()
    print(f"All lock states: {all_states}")

    assert all_states["LEFT"] is True
    assert isinstance(all_states["RIGHT"], bool)

    print("Unlocking LEFT button...")
    makcu.lock_left(False)

    print("Rechecking LEFT lock state after unlock...")
    assert not makcu.is_button_locked(MouseButton.LEFT)

def test_makcu_behavior(makcu):
    makcu.move(25, 25)
    makcu.click(MouseButton.LEFT)
    makcu.scroll(-2)

def test_reset_all(makcu):
    makcu.mouse.lock_left(False)
    makcu.mouse.lock_right(False)
    makcu.mouse.lock_middle(False)
    makcu.mouse.lock_side1(False)
    makcu.mouse.lock_side2(False)
    makcu.mouse.lock_x(False)
    makcu.mouse.lock_y(False)

    states = makcu.mouse.get_all_lock_states()
    assert all(state is False for state in states.values() if state is not None), \
        f"Expected all unlocked, got: {states}"

    makcu.enable_button_monitoring(False)


from .controller import MakcuController
from .enums import MouseButton
from .errors import MakcuError, MakcuConnectionError

def create_controller(fallback_com_port="", debug=False, send_init=True):
    makcu = MakcuController(fallback_com_port, debug=debug, send_init=send_init)
    makcu.connect()
    return makcu

__all__ = [
    "MakcuController",
    "MouseButton",
    "MakcuError",
    "MakcuConnectionError",
    "create_controller",
]


import sys
import webbrowser
import os
from pathlib import Path
import pytest
from makcu import create_controller, MakcuConnectionError

def debug_console():
    controller = create_controller()
    transport = controller.transport

    print("🔧 Makcu Debug Console")
    print("Type a raw command (e.g., km.version()) and press Enter.")
    print("Type 'exit' or 'quit' to leave.")

    while True:
        try:
            cmd = input(">>> ").strip()
            if cmd.lower() in {"exit", "quit"}:
                break
            if not cmd:
                continue

            response = transport.send_command(cmd, expect_response=True)
            print(f"{response or '(no response)'}")

        except Exception as e:
            print(f"⚠️ Error: {e}")

    controller.disconnect()
    print("Disconnected.")

def test_port(port):
    try:
        print(f"Trying to connect to {port} (without init command)...")
        controller = create_controller(send_init=False)
        print(f"✅ Successfully connected to {port}")
        controller.disconnect()
    except MakcuConnectionError as e:
        print(f"❌ Failed to connect to {port}: {e}")
    except Exception as e:
        print(f"❌ Unexpected error: {e}")

def run_tests():
    print("🧪 Running Pytest Suite...")

    package_dir = Path(__file__).resolve().parent
    test_file = package_dir / "test_suite.py"

    sys.exit(pytest.main([
        str(test_file),
        "--rootdir", str(package_dir),
        "-v", "--tb=short",
        "--capture=tee-sys",
        "--html=latest_pytest.html",
        "--self-contained-html"
    ]))

    report_path = os.path.abspath("latest_pytest.html")
    if os.path.exists(report_path):
        print(f"📄 Opening test report: {report_path}")
        webbrowser.open(f"file://{report_path}")
    else:
        print("❌ Report not found. Something went wrong.")

    if result.returncode != 0:
        print("❌ Some tests failed.")
    else:
        print("✅ All tests passed.")

def main():
    args = sys.argv[1:]

    if not args:
        print("Usage:")
        print("  python -m makcu --debug")
        print("  python -m makcu --testPort COM3")
        print("  python -m makcu --runtest")
        return

    if args[0] == "--debug":
        debug_console()
    elif args[0] == "--testPort" and len(args) == 2:
        test_port(args[1])
    elif args[0] == "--runtest":
        run_tests()
    else:
        print(f"Unknown command: {' '.join(args)}")
        print("Usage:")
        print("  python -m makcu --debug")
        print("  python -m makcu --testPort COM3")
        print("  python -m makcu --runtest")

if __name__ == "__main__":
    main()


import serial
import threading
import time
from serial.tools import list_ports
from .errors import MakcuConnectionError, MakcuTimeoutError
from .enums import MouseButton

class SerialTransport:
    baud_change_command = bytearray([0xDE, 0xAD, 0x05, 0x00, 0xA5, 0x00, 0x09, 0x3D, 0x00])

    button_map = {
        0: 'left',
        1: 'right',
        2: 'middle',
        3: 'mouse4',
        4: 'mouse5'
    }

    def __init__(self, fallback, debug=False, send_init=True):        
        self._fallback_com_port = fallback
        self._log_messages = []
        self.debug = debug
        self.send_init = send_init
        self._button_callback = None
        self._last_mask = 0
        self._lock = threading.Lock()
        self._is_connected = False
        self._stop_event = threading.Event()
        self._listener_thread = None
        self._button_states = {btn: False for btn in self.button_map.values()}
        self._last_callback_time = {bit: 0 for bit in self.button_map}
        self._pause_listener = False

        self._button_enum_map = {
            0: MouseButton.LEFT,
            1: MouseButton.RIGHT,
            2: MouseButton.MIDDLE,
            3: MouseButton.MOUSE4,
            4: MouseButton.MOUSE5,
        }

        self.port = self.find_com_port()
        if not self.port:
            raise MakcuConnectionError("Makcu device not found. Please specify a port explicitly.")

        self.baudrate = 115200
        self.serial = None
        self._current_baud = None


    def receive_response(self, max_bytes=1024, max_lines=3, sent_command: str = "") -> str:
        lines = []
        try:
            for _ in range(max_lines):
                line = self.serial.readline(max_bytes)
                if not line:
                    break
                decoded = line.decode(errors="ignore").strip()
                if decoded:
                    lines.append(decoded)
        except Exception as e:
            print(f"[RECV ERROR] {e}")
            return ""

        command_clean = sent_command.strip()
        if lines:
            lines.pop(-1)
        if command_clean in lines and len(lines) > 1:
            lines.remove(command_clean)
        return "\n".join(lines)

    def set_button_callback(self, callback):
        self._button_callback = callback

    def _log(self, message):
        timestamp = time.strftime("%H:%M:%S")
        entry = f"[{timestamp}] {message}"
        self._log_messages.append(entry)
        if len(self._log_messages) > 20:
            self._log_messages.pop(0)
        print(entry, flush=True)

    def find_com_port(self):
        self._log("Searching for CH343 device...")

        for port in list_ports.comports():
            if "VID:PID=1A86:55D3" in port.hwid.upper():
                self._log(f"Device found: {port.device}")
                return port.device

        if self._fallback_com_port:
            self._log(f"Device not found. Falling back to specified port: {self._fallback_com_port}")
            return self._fallback_com_port
        else:
            self._log("Fallback port not specified or invalid.")
            return None

    def _open_serial_port(self, port, baud_rate):
        try:
            self._log(f"Trying to open {port} at {baud_rate} baud.")
            return serial.Serial(port, baud_rate, timeout=0.05)
        except serial.SerialException:
            self._log(f"Failed to open {port} at {baud_rate} baud.")
            return None

    def _change_baud_to_4M(self):
        if self.serial and self.serial.is_open:
            self._log("Sending baud rate switch command to 4M.")
            self.serial.write(self.baud_change_command)
            self.serial.flush()
            time.sleep(0.05)
            self.serial.baudrate = 4000000
            self._current_baud = 4000000
            self._log("Switched to 4M baud successfully.")
            return True
        return False


    def connect(self):
        if self._is_connected:
            self._log("Already connected.")
            return
        self.serial = self._open_serial_port(self.port, 115200)
        if not self.serial:
            raise MakcuConnectionError(f"Failed to connect to {self.port} at 115200.")
        self._log(f"Connected to {self.port} at 115200.")
        if not self._change_baud_to_4M():
            raise MakcuConnectionError("Failed to switch to 4M baud.")
        self._is_connected = True
        if self.send_init:
            with self._lock:
                self.serial.write(b"km.buttons(1)\r")
                self.serial.flush()
                self._log("Sent init command: km.buttons(1)")

            self._stop_event.clear()
            self._listener_thread = threading.Thread(target=self._listen, kwargs={"debug": self.debug}, daemon=True)
            self._listener_thread.start()

    def disconnect(self):
        if self.send_init:
            self._stop_event.set()
            if self._listener_thread:
                self._listener_thread.join()
        with self._lock:
            if self.serial and self.serial.is_open:
                self.serial.close()
            self.serial = None
            self._is_connected = False
            self._log("Disconnected.")

    def is_connected(self):
        return self._is_connected

    def send_command(self, command, expect_response=False):
        time.sleep(0.06)
        if not self._is_connected or not self.serial or not self.serial.is_open:
            raise MakcuConnectionError("Serial connection not open.")
        with self._lock:
            try:
                self._pause_listener = True
                self.serial.reset_input_buffer()
                self.serial.write(command.encode("ascii") + b"\r\n")
                self.serial.flush()
                if expect_response:
                    response = self.receive_response(sent_command=command)
                    if not response:
                        raise MakcuTimeoutError(f"No response from device for command: {command}")
                    return response
            finally:
                self._pause_listener = False

    def get_button_states(self):
        return dict(self._button_states)

    def get_button_mask(self) -> int:
        mask = 0
        for i, name in self.button_map.items():
            if self._button_states.get(name, False):
                mask |= (1 << i)
        return mask

    def enable_button_monitoring(self, enable: bool = True):
        self.send_command("km.buttons(1)" if enable else "km.buttons(0)")

    def catch_button(self, button: str):
        command = {
            "LEFT": "km.catch_ml(0)",
            "RIGHT": "km.catch_mr(0)",
            "MIDDLE": "km.catch_mm(0)",
            "MOUSE4": "km.catch_ms1(0)",
            "MOUSE5": "km.catch_ms2(0)",
        }.get(button.upper())
        if command:
            self.send_command(command)
        else:
            raise ValueError(f"Unsupported button: {button}")

    def read_captured_clicks(self, button: str) -> int:
        command = {
            "LEFT": "km.catch_ml()",
            "RIGHT": "km.catch_mr()",
            "MIDDLE": "km.catch_mm()",
            "MOUSE4": "km.catch_ms1()",
            "MOUSE5": "km.catch_ms2()",
        }.get(button.upper())
        if command:
            result = self.send_command(command, expect_response=True)
            try:
                return int(result.strip())
            except Exception:
                return 0
        else:
            raise ValueError(f"Unsupported button: {button}")

    def _listen(self, debug=False):
        self._log("Started listener thread")
        button_states = {i: False for i in self.button_map}
        self._last_mask = 0
        self._last_callback_time = {bit: 0 for bit in self.button_map}

        while self._is_connected and not self._stop_event.is_set():
            if self._pause_listener:
                time.sleep(0.001)
                continue

            try:
                byte = self.serial.read(1)
                if byte:
                    value = byte[0]

                    if value != self._last_mask:
                        byte_str = str(byte)

                        if "b'\\x00" in byte_str:
                            for bit in self.button_map:
                                button_states[bit] = False

                        elif "b'\\x" in byte_str:
                            for bit, name in self.button_map.items():
                                is_pressed = bool(value & (1 << bit))
                                if is_pressed != button_states[bit]:
                                    button_states[bit] = is_pressed

                        for bit, name in self.button_map.items():
                            self._button_states[name] = button_states[bit]

                        if self._button_callback:
                            for bit, name in self.button_map.items():
                                previous = bool(self._last_mask & (1 << bit))
                                current = bool(value & (1 << bit))
                                if previous != current:
                                    button_enum = self._button_enum_map.get(bit)
                                    if button_enum:
                                        self._button_callback(button_enum, current)

                        self._last_mask = value

                        if debug:
                            pressed = [name for bit, name in self.button_map.items() if button_states[bit]]
                            button_str = ", ".join(pressed) if pressed else "No buttons pressed"
                            self._log(f"Byte: {value} (0x{value:02X}) -> {button_str}")

            except serial.SerialException as e:
                if "ClearCommError failed" not in str(e):
                    self._log(f"Serial error during listening: {e}")
                    break

            time.sleep(0.0001)

        self._log("Listener thread exiting")


class MakcuError(Exception):
    """Base exception for all Makcu-related errors."""
    pass

class MakcuConnectionError(MakcuError):
    """Raised when the device connection fails."""
    pass

class MakcuCommandError(MakcuError):
    """Raised when a device command is invalid, rejected, or fails."""
    pass

class MakcuTimeoutError(MakcuError):
    """Raised when the device does not respond in time."""
    pass

class MakcuResponseError(MakcuError):
    """Raised when the response from the device is malformed or unexpected."""
    pass